#!/usr/bin/env python

from __future__ import print_function

import cStringIO
import os
import os.path
import subprocess
import tempfile

import util


# 
# Utility Functions
# 

def _check_compression(filename, *args, **kwargs):
	with util.archive_created(filename, "/bin", *args, **kwargs) as path:
		with util.directory_created("extracted") as directory:
			subprocess.check_call(["xar", "-x", "-f", path, "-C", directory])
			util.assert_identical_directories("/bin", os.path.join(directory, "bin"))


# 
# Test Cases
# 

def no_compression(filename):
	_check_compression(filename, "--compression=none")

def default_compression(filename):
	_check_compression(filename)

def gzip_compression_long(filename):
	util.skip_if_no_compression_support("gzip")
	_check_compression(filename, "--compression=gzip")

def gzip_compression_short(filename):
	util.skip_if_no_compression_support("gzip")
	_check_compression(filename, "-z")

def bzip2_compression_long(filename):
	util.skip_if_no_compression_support("bzip2")
	_check_compression(filename, "--compression=bzip2")

def bzip2_compression_short(filename):
	util.skip_if_no_compression_support("bzip2")
	_check_compression(filename, "-j")

def lzma_compression_long(filename):
	util.skip_if_no_compression_support("lzma")
	_check_compression(filename, "--compression=lzma")

def lzma_compression_short(filename):
	util.skip_if_no_compression_support("lzma")
	_check_compression(filename, "-a")

TEST_CASES = (no_compression, default_compression,
              gzip_compression_long, bzip2_compression_long, lzma_compression_long,
              gzip_compression_short, bzip2_compression_short, lzma_compression_short)

if __name__ == "__main__":
	for case in TEST_CASES:
		try:
			case("{f}.xar".format(f=case.func_name))
			print("PASSED: {f}".format(f=case.func_name))
		except (AssertionError, IOError, subprocess.CalledProcessError):
			import sys, os
			print("FAILED: {f}".format(f=case.func_name))
			sys.excepthook(*sys.exc_info())
			print("")
		except util.TestCaseSkipError, e:
			print("SKIPPED: {f}: {m}".format(f=case.func_name, m=e.message))



#!/usr/bin/env python

from __future__ import print_function

import os
import os.path
import subprocess

import util

# 
# Utility Functions
# 

def _assert_same_inodes(path1, path2):
	assert os.path.exists(path1), "missing hard-linked-file {p}".format(p=path1)
	assert os.path.exists(path1), "missing hard-linked-file {p}".format(p=path2)
	stat1 = os.lstat(path1)
	stat2 = os.lstat(path2)
	assert stat1.st_ino == stat2.st_ino, "inodes do not match for hard-linked files \"{p1}\" and \"{p2}\"".format(p1=path1, p2=path2)


# 
# Test Cases
# 

def hard_link_in_directory(filename):
	with util.directory_created("hardlink_scratch") as directory:
		util.touch(os.path.join(directory, "a"))
		os.link(os.path.join(directory, "a"), os.path.join(directory, "b"))
		
		with util.archive_created("hardlink_archive.xar", "hardlink_scratch") as path:
			with util.directory_created("hardlink_extracted") as extracted:
				subprocess.check_call(["xar", "-x", "-C", extracted, "-f", path])
				_assert_same_inodes(os.path.join(extracted, "hardlink_scratch", "a"), os.path.join(extracted, "hardlink_scratch", "b"))

def hard_link_in_cwd(filename):
	with util.directory_created("hardlink_scratch") as directory:
		with util.chdir(directory):
			util.touch("a")
			os.link("a", "b")
			with util.archive_created(os.path.join("..", "hardlink_archive.xar"), ".") as path:
				with util.directory_created(os.path.join("..", "hardlink_extracted")) as extracted:
					subprocess.check_call(["xar", "-x", "-C", extracted, "-f", path])
					_assert_same_inodes(os.path.join(extracted, "a"), os.path.join(extracted, "b"))

def hard_link_identical_files(filename):
	with util.directory_created("hardlink_scratch") as directory:
		with open(os.path.join(directory, "a"), "w") as f:
			f.write("1234samecontent")
		with open(os.path.join(directory, "b"), "w") as f:
			f.write("1234samecontent")
		with util.archive_created("hardlink_archive.xar", "hardlink_scratch", "--link-same") as path:
			with util.directory_created("hardlink_extracted") as extracted:
				subprocess.check_call(["xar", "-x", "-C", extracted, "-f", path])
				_assert_same_inodes(os.path.join(extracted, "hardlink_scratch", "a"), os.path.join(extracted, "hardlink_scratch", "b"))

TEST_CASES = (hard_link_in_directory, hard_link_in_cwd, hard_link_identical_files)

if __name__ == "__main__":
	for case in TEST_CASES:
		try:
			case("{f}.xar".format(f=case.func_name))
			print("PASSED: {f}".format(f=case.func_name))
		except (AssertionError, IOError, subprocess.CalledProcessError):
			import sys, os
			print("FAILED: {f}".format(f=case.func_name))
			sys.excepthook(*sys.exc_info())
			print("")



#!/usr/bin/env python

import contextlib
import hashlib
import os
import os.path
import shutil
import stat
import subprocess
import sys
import xattr

class TestCaseSkipError(Exception):
	pass

def skip_if_no_compression_support(type):
	"""
	Raises TestCaseSkipError if the type is "lzma" and the test is running on
	darwin (OS X). In the future, we should add a hidden debugging flag to xar
	to determine valid compression types. This will skip incorrectly if a
	custom xar is used on OS X, or if a custom xar on another platform is
	built without bzip2 or lzma.
	
	"""
	if sys.platform == "darwin" and type == "lzma":
		raise TestCaseSkipError("{t} support not compiled in".format(t=type))

@contextlib.contextmanager
def directory_created(directory_path):
	"""
	Creates the named directory and provides the path to the directory to the
	calling code. Automatically removes the directory when finished.
	
	Usage:
	with directory_created("foobar") as path:
		do_stuff_with_path
	
	"""
	os.mkdir(directory_path)
	try:
		yield os.path.realpath(directory_path)
	finally:
		if os.path.exists(directory_path):
			shutil.rmtree(directory_path)

@contextlib.contextmanager
def archive_created(archive_path, content_path, *extra_args, **extra_kwargs):
	"""
	Creates a named xar archive of the specified content path, returning the
	path to the archive. Automatically removes the archive when finished.
	
	Usage:
	with archive_created("/bin", "bin.xar") as path:
		do_stuff_with(path)
	
	"""
	cmd = ["xar", "-c", "-f", archive_path, content_path]
	if extra_args:
		cmd += list(extra_args)
	try:
		subprocess.check_call(cmd, **extra_kwargs)
		assert os.path.exists(archive_path), "failed to create archive \"{p}\" but xar did not report an error".format(p=archive_path)
		yield os.path.realpath(archive_path)
	finally:
		if os.path.exists(archive_path):
			os.unlink(archive_path)

HASH_CHUNK_SIZE = 32768

def _md5_path(path):
	with open(path, "r") as f:
		h = hashlib.md5()
		while True:
			last = f.read(HASH_CHUNK_SIZE)
			if not last:
				break
			h.update(last)
		return h.digest()

def assert_identical_directories(path1, path2):
	"""
	Verifies two directories have identical contents. Checks file type (via
	the high byte of the mode), size, atime, and mtime, but does not check
	other attributes like uid and gid, since they can be expected to change.
	
	"""
	seen = set([])
	for file1 in os.listdir(path1):
		seen.add(file1)
		entry1 = os.path.join(path1, file1)
		entry2 = os.path.join(path2, file1)
		assert os.path.exists(entry2), "\"{f1}\" exists in \"{p1}\" but not \"{p2}\"".format(f1=file1, p1=path1, p2=path2)
		
		# Extended attributes
		xattr1 = xattr.xattr(entry1)
		xattr2 = xattr.xattr(entry2)
		assert set(xattr1.list()) == set(xattr2.list()), "list of extended attributes on \"{f1}\" ({l1}) differs from \"{f2}\" ({l2})".format(f1=entry1, l1=xattr1.list(), f2=entry2, l2=xattr2.list())
		for attribute in xattr1.list():
			assert xattr1.get(attribute) == xattr2.get(attribute), "extended attribute \"{a1}\" on \"{f1}\" doesn't match value from \"{f2}\"".format(a1=attribute, f1=entry1, f2=entry2)
		
		# Why do it this way? We want to lstat() instead of stat(), so we can't use os.path.isdir() and friends
		stat1 = os.lstat(entry1)
		stat2 = os.lstat(entry2)
		
		# Modes
		mode1 = stat1.st_mode
		mode2 = stat2.st_mode
		if stat.S_ISREG(mode1):
			assert stat.S_ISREG(mode2)
		if stat.S_ISDIR(mode1):
			assert stat.S_ISDIR(mode2)
		if stat.S_ISLNK(mode1):
			assert stat.S_ISLNK(mode2)
		if stat.S_ISCHR(mode1):
			assert stat.S_ISCHR(mode2)
		if stat.S_ISBLK(mode1):
			assert stat.S_ISBLK(mode2)
		if stat.S_ISFIFO(mode1):
			assert stat.S_ISFIFO(mode2)
		if stat.S_ISSOCK(mode1):
			assert stat.S_ISSOCK(mode2)
		
		# Sizes and the like
		assert stat1.st_size == stat2.st_size, "size mismatch for \"{e1}\" ({s1}) and \"{e2}\" ({s2})".format(e1=entry1, s1=stat1.st_size, e2=entry2, s2=stat2.st_size)
		assert stat1.st_mtime == stat2.st_mtime, "mtime mismatch for \"{e1}\" and \"{e2}\"".format(e1=entry1, e2=entry2)
		assert _md5_path(entry1) == _md5_path(entry2), "md5 hash mismatch for \"{e1}\" and \"{e2}\"".format(e1=entry1, e2=entry2)
		if os.path.isdir(entry1):
			assert_identical_directories(entry1, entry2)
	for file2 in os.listdir(path2):
		assert file2 in seen, "\"{f2}\" exists in \"{p2}\" but not \"{p1}\"".format(f2=file2, p1=path1, p2=path2)

def touch(path):
	if not os.path.exists(path):
		with open(path, "w"):
			pass
	os.utime(path, None)

@contextlib.contextmanager
def chdir(*args, **kwargs):
	cwd = os.getcwd()
	os.chdir(*args, **kwargs)
	try:
		yield os.getcwd()
	finally:
		os.chdir(cwd)



#!/usr/bin/env python

from __future__ import print_function

import os
import os.path
import subprocess

import util

# 
# Utility Functions
# 

def _test_truncation(filename, path_to_be_archived, bytes_to_chop, *args):
	with util.archive_created(filename, path_to_be_archived) as path:
		with open("/dev/null", "w") as bitbucket:
			size = os.stat(path).st_size
			while size > 0:
				last_size = size
				size = max(0, size - bytes_to_chop)
				with open(path, "w+") as f:
					f.truncate(size)
			
				with util.directory_created("scratch") as directory:
					returncode = subprocess.call(["xar", "-x", "-f", path, "-C", directory], stderr=bitbucket)
					assert returncode != 0, "xar claimed to succeed when extracting a truncated archive"

# 
# Test Cases
#

def large_uncompressed(filename):
	_test_truncation(filename, "/usr/share/man/man1", 1024 * 1024, "--compression=none")
	
def large_default_compression(filename):
	_test_truncation(filename, "/usr/share/man/man1", 1024 * 1024)

def large_gzip_compressed(filename):
	util.skip_if_no_compression_support("gzip")
	_test_truncation(filename, "/usr/share/man/man1", 1024 * 1024, "--compression=gzip")

def large_bzip2_compressed(filename):
	util.skip_if_no_compression_support("bzip2")
	_test_truncation(filename, "/usr/share/man/man1", 1024 * 1024, "--compression=bzip2")

def large_lzma_compressed(filename):
	util.skip_if_no_compression_support("lzma")
	_test_truncation(filename, "/usr/share/man/man1", 1024 * 1024, "--compression=lzma")

# "small" variants use a non-base-2 size to try to catch issues that occur on uneven boundaries

def small_uncompressed(filename):
	_test_truncation(filename, "/bin", 43651, "--compression=none")

def small_default_compression(filename):
	_test_truncation(filename, "/bin", 43651)

def small_gzip_compressed(filename):
	util.skip_if_no_compression_support("gzip")
	_test_truncation(filename, "/bin", 43651, "--compression=gzip")

def small_bzip2_compressed(filename):
	util.skip_if_no_compression_support("bzip2")
	_test_truncation(filename, "/bin", 43651, "--compression=bzip2")

def small_lzma_compressed(filename):
	util.skip_if_no_compression_support("lzma")
	_test_truncation(filename, "/bin", 43651, "--compression=lzma")


TEST_CASES = (large_uncompressed, large_default_compression,
              large_gzip_compressed, large_bzip2_compressed, large_lzma_compressed,
              small_uncompressed, small_default_compression,
              small_gzip_compressed, small_bzip2_compressed, small_lzma_compressed)

if __name__ == "__main__":
	for case in TEST_CASES:
		try:
			case("{f}.xar".format(f=case.func_name))
			print("PASSED: {f}".format(f=case.func_name))
		except (AssertionError, IOError, subprocess.CalledProcessError):
			import sys, os
			print("FAILED: {f}".format(f=case.func_name))
			sys.excepthook(*sys.exc_info())
			print("")
		except util.TestCaseSkipError, e:
			print("SKIPPED: {f}: {m}".format(f=case.func_name, m=e.message))



#!/usr/bin/env python

from __future__ import print_function

import glob
import os
import os.path
import re
import shutil
import subprocess

import util


# 
# Utility Functions
# 

def _file_offsets_for_archive(path, xsl_path):
	subprocess.check_call(["xar", "--dump-toc=heap_toc.xml", "-f", path])
	try:
		offsets = subprocess.check_output(["xsltproc", xsl_path, "heap_toc.xml"])
		matches = [re.match("^(.+)\s([^\s]+)$", offset) for offset in offsets.splitlines()]
		offsets = [(match.groups()[0], int(match.groups()[1])) for match in matches]
		return offsets
	finally:
		os.unlink("heap_toc.xml")

# 
# Test Cases
# 

XSL_PATH = os.path.join(os.path.dirname(os.path.realpath(__file__)), "heap1.xsl")

def normal_heap(filename):
	with util.directory_created("scratch") as directory:
		shutil.copy("/bin/ls", os.path.join(directory, "ls"))
		shutil.copy(os.path.join(directory, "ls"), os.path.join(directory, "foo"))
		with util.chdir(directory):
			with util.archive_created(os.path.join("..", "heap.xar"), ".") as path:
				# Verify file offsets are as we expect
				offsets = _file_offsets_for_archive(path, XSL_PATH)
				(f1, o1) = offsets[0]
				(f2, o2) = offsets[1]
				
				assert o1 < o2, "offset for first file \"{f1}\" ({o1}) greater than or equal to offset for last file \"{f2}\" ({o2})".format(f1=f1, o1=o1, f2=f2, o2=o2)
				
				# Make sure extraction goes all right
				with util.directory_created("extracted") as extracted:
					subprocess.check_call(["xar", "-x", "-f", path, "-C", extracted])

def coalesce_heap(filename):
	with util.directory_created("scratch") as directory:
		shutil.copy("/bin/ls", os.path.join(directory, "ls"))
		shutil.copy(os.path.join(directory, "ls"), os.path.join(directory, "foo"))
		with util.chdir(directory):
			with util.archive_created(os.path.join("..", "heap.xar"), ".", "--coalesce-heap") as path:
				# Verify file offsets are as we expect
				offsets = _file_offsets_for_archive(path, XSL_PATH)
				(f1, o1) = offsets[0]
				(f2, o2) = offsets[1]
				
				# Make sure extraction goes all right
				with util.directory_created("extracted") as extracted:
					subprocess.check_call(["xar", "-x", "-f", path, "-C", extracted])

TEST_CASES = (normal_heap, coalesce_heap)

if __name__ == "__main__":
	for case in TEST_CASES:
		try:
			case("{f}.xar".format(f=case.func_name))
			print("PASSED: {f}".format(f=case.func_name))
		except (AssertionError, IOError, subprocess.CalledProcessError):
			import sys, os
			print("FAILED: {f}".format(f=case.func_name))
			sys.excepthook(*sys.exc_info())
			print("")



#!/usr/bin/env python

from __future__ import print_function

import os
import os.path
import shutil
import subprocess
import xattr

import util

# Notes:
# * Due to internal buffer sizing, "large" should be at least 4096 bytes for
#   uncompressed archives and archives with individual file compression, and
#   and at least 32768 bytes for archives with compressed heaps.
# * For large files, we intentionally use a non-base-2 file size to ensure we
#   catch file extraction bugs for files with a size not equal to the internal
#   read buffer size (which is base-2).
# * For large extended attributes, we do the same thing.

# 
# Utility Functions
# 

class MissingExtendedAttributeError(AssertionError):
	pass

def _random_big_data(bytes=65536, path="/dev/random"):
	"""
	Returns a random string with the number of bytes requested. Due to xar
	implementation details, this should be greater than 4096 (32768 for
	compressed heap testing).
	
	"""
	with open(path, "r") as f:
		return f.read(bytes)

def _test_xattr_on_file_with_contents(filename, file_contents, xattrs=[], xar_create_flags=[], xar_extract_flags=[]):
	try:
		# Write file out
		with open(filename, "w") as f:
			f.write(file_contents)
			for (key, value) in xattrs:
				xattr.setxattr(f, key, value)
		
		# Store it into a xarchive
		archive_name = "{f}.xar".format(f=filename)
		with util.archive_created(archive_name, filename, *xar_create_flags) as path:
			# Extract xarchive
			with util.directory_created("extracted") as directory:
				# Validate resulting xattrs
				subprocess.check_call(["xar", "-x", "-C", directory, "-f", path] + xar_extract_flags)
				for (key, value) in xattrs:
					try:
						assert xattr.getxattr(os.path.join(directory, filename), key) == value, "extended attribute \"{n}\" has incorrect contents after extraction".format(n=key)
					except KeyError:
						raise MissingExtendedAttributeError("extended attribute \"{n}\" missing after extraction".format(n=key))
				
				# Validate file contents
				with open(os.path.join(directory, filename), "r") as f:
					if f.read() != file_contents:
						raise MissingExtendedAttributeError("archived file \"{f}\" has has incorrect contents after extraction".format(f=filename))
	finally:
		os.unlink(filename)


# 
# Test Cases
# 

# Note: xar currently drops empty extended attributes (and any xar_prop_t with empty contents, actually). The empty
#       tests are commented out awaiting a day when this might be different.

# def empty_xattr_empty_file(filename):
# 	_test_xattr_on_file_with_contents(filename, "", xattrs=[("foo", "")])

def small_xattr_empty_file(filename):
	_test_xattr_on_file_with_contents(filename, "", xattrs=[("foo", "1234")])

def large_xattr_empty_file(filename):
	_test_xattr_on_file_with_contents(filename, "", xattrs=[("foo", _random_big_data(5000))])

# def empty_xattr_small_file(filename):
# 	_test_xattr_on_file_with_contents(filename, "small.file.contents", xattrs=[("foo", "")])

def small_xattr_small_file(filename):
	_test_xattr_on_file_with_contents(filename, "small.file.contents", xattrs=[("foo", "1234")])

def large_xattr_small_file(filename):
	_test_xattr_on_file_with_contents(filename, "small.file.contents", xattrs=[("foo", _random_big_data(4567))])

# def empty_xattr_large_file(filename):
# 	_test_xattr_on_file_with_contents(filename, _random_big_data(10000000), xattrs=[("foo", "")])

def small_xattr_large_file(filename):
	_test_xattr_on_file_with_contents(filename, _random_big_data(5000000), xattrs=[("foo", "1234")])

def large_xattr_large_file(filename):
	_test_xattr_on_file_with_contents(filename, _random_big_data(9876543), xattrs=[("foo", _random_big_data(6543))])

def multiple_xattrs(filename):
	_test_xattr_on_file_with_contents(filename, "", xattrs=[("foo", "bar"), ("baz", "1234"), ("quux", "more")]) # ("empty", "")

def distribution_create(filename):
	try:
		_test_xattr_on_file_with_contents(filename, "dummy", xattrs=[("foo", "bar")], xar_create_flags=["--distribution"])
	except MissingExtendedAttributeError:
		pass
	else:
		raise AssertionError("no error encountered")

# Note: xar currently has no way to exclude a property on extraction. This test is commented out awaiting the day
#       when it can.

# def distribution_extract(filename):
# 	try:
# 		_test_xattr_on_file_with_contents(filename, "dummy", xattrs=[("foo", "bar")], xar_extract_flags=["--distribution"])
# 	except MissingExtendedAttributeError:
# 		pass
# 	else:
# 		raise AssertionError("no error encountered")


TEST_CASES = (small_xattr_empty_file, large_xattr_empty_file,
              small_xattr_small_file, large_xattr_small_file,
              small_xattr_large_file, large_xattr_large_file,
              multiple_xattrs, distribution_create)

if __name__ == "__main__":
	for case in TEST_CASES:
		try:
			case(case.func_name)
			print("PASSED: {f}".format(f=case.func_name))
		except (AssertionError, IOError, subprocess.CalledProcessError):
			import sys, os
			print("FAILED: {f}".format(f=case.func_name))
			sys.excepthook(*sys.exc_info())
			print("")



#!/usr/bin/env python

from __future__ import print_function

import contextlib
import os
import os.path
import subprocess
import util


# 
# Utility Functions
# 

@contextlib.contextmanager
def _data_test(filename, *args, **kwargs):
	with util.directory_created("data_scratch") as directory:
		yield directory
		# Files are now created
		with util.archive_created("data.xar", "./data_scratch", *args, **kwargs) as path:
			_process_toc(path)
			with util.directory_created("data_extracted") as extracted:
				subprocess.check_call(["xar", "-x", "-f", path, "-C", extracted])
				util.assert_identical_directories(directory, os.path.join(extracted, "data_scratch"))

def _process_toc(archive_path):
	subprocess.check_call(["xar", "-f", archive_path, "--dump-toc=data_toc.xml"])
	try:
		result = subprocess.check_output(["xsltproc", "-o", "-", os.path.realpath(os.path.join(__file__, "..", "data.xsl")), "data_toc.xml"])
		assert result == "", "expected no data offset, but instead found:{o}".format(o=result)
	finally:
		os.unlink("data_toc.xml")

def _zero_length(filename, *args, **kwargs):
	with _data_test(filename, *args, **kwargs) as directory:
		util.touch(os.path.join(directory, "empty"))


# 
# Test Cases
# 

def zero_length_default_compression(filename):
	_zero_length(filename)

def zero_length_no_compression(filename):
	_zero_length(filename, "--compression=none")

def zero_length_gzip_compression(filename):
	util.skip_if_no_compression_support("gzip")
	_zero_length(filename, "--compression=gzip")

def zero_length_bzip2_compression(filename):
	util.skip_if_no_compression_support("bzip2")
	_zero_length(filename, "--compression=bzip2")

def zero_length_lzma_compression(filename):
	util.skip_if_no_compression_support("lzma")
	_zero_length(filename, "--compression=lzma")

def _mixed_length(filename, *args, **kwargs):
	with _data_test(filename, *args, **kwargs) as directory:
		util.touch(os.path.join(directory, "mixed_empty"))
		with open(os.path.join(directory, "mixed_small"), "w") as f:
			f.write("1234")

def mixed_length_no_compression(filename):
	_mixed_length(filename, "--compression=none")

def mixed_length_default_compression(filename):
	_mixed_length(filename)

def mixed_length_gzip_compression(filename):
	util.skip_if_no_compression_support("gzip")
	_mixed_length(filename, "--compression=gzip")

def mixed_length_bzip2_compression(filename):
	util.skip_if_no_compression_support("bzip2")
	_mixed_length(filename, "--compression=bzip2")

def mixed_length_lzma_compression(filename):
	util.skip_if_no_compression_support("lzma")
	_mixed_length(filename, "--compression=lzma")


TEST_CASES = (zero_length_default_compression, zero_length_no_compression,
              zero_length_gzip_compression, zero_length_bzip2_compression, zero_length_lzma_compression,
              mixed_length_no_compression, mixed_length_default_compression,
              mixed_length_gzip_compression, mixed_length_bzip2_compression, mixed_length_lzma_compression)

if __name__ == "__main__":
	for case in TEST_CASES:
		try:
			case("{f}.xar".format(f=case.func_name))
			print("PASSED: {f}".format(f=case.func_name))
		except (AssertionError, IOError, subprocess.CalledProcessError):
			import sys, os
			print("FAILED: {f}".format(f=case.func_name))
			sys.excepthook(*sys.exc_info())
			print("")
		except util.TestCaseSkipError, e:
			print("SKIPPED: {f}: {m}".format(f=case.func_name, m=e.message))



#!/usr/bin/env python

from __future__ import print_function

import hashlib
import os
import os.path
import re
import shutil
import struct
import subprocess

import util


# 
# Utility Functions
# 

def _get_numeric_value_from_header(archive_name, key):
	"""
	Dumps the header of the specified xar archive and extracts the header
	size from the output, in bytes.
	
	"""
	header = subprocess.check_output(["xar", "--dump-header", "-f", archive_name])
	for line in header.splitlines():
		matchdata = re.match("^(.+):\s+(.+)$", line) # magic: 0x78617221 (OK)
		assert matchdata, "unexpected output from `xar --dump-header`:\n{h}".format(h=header)
		if matchdata.groups()[0] == key:
			return int(matchdata.groups()[1])
	raise AssertionError("no \"{k}\" found for archive \"{n}\":\n{h}".format(k=key, n=archive_name, h=header))

def _get_header_size(archive_name):
	return _get_numeric_value_from_header(archive_name, "size")

def _get_toc_size(archive_name):
	return _get_numeric_value_from_header(archive_name, "Compressed TOC length")

def _clobber_bytes_at(clobber_range, path):
	with open(path, "r+") as f:
		f.seek(clobber_range[0])
		with open("/dev/random", "r") as r:
			random_bytes = r.read(len(clobber_range))
			f.write(random_bytes)

def _verify_extraction_failed(filename):
	with util.directory_created("extracted") as directory:
		try:
			with open("/dev/null", "w") as n:
				returncode = subprocess.call(["xar", "-x", "-C", directory, "-f", filename], stdout=n, stderr=n)
			assert returncode != 0, "xar reported success extracting an archive with a broken TOC"
		finally:
			if os.path.exists(directory):
				shutil.rmtree(directory)

def _hex_digest_string(raw_digest):
	unpack_string = "B" * len(raw_digest)
	format_string = "%02x" * len(raw_digest)
	return format_string % struct.unpack(unpack_string, raw_digest)

def _verify_header_checksum(filename, algorithm):
	header_size = _get_header_size(filename)
	toc_length = _get_toc_size(filename)
	
	with open(filename, "r") as f:
		f.seek(header_size)
		h = hashlib.new(algorithm, f.read(toc_length))
		computed_digest = h.digest()
		# We intentionally ignore the TOC-specified offset. We should build a variant of this that uses the TOC
		# offset so we check both.
		stored_digest = f.read(len(computed_digest))
		assert computed_digest == stored_digest, "Digests don't match: expected value {d1} != stored value {d2}".format(d1=_hex_digest_string(computed_digest), d2=_hex_digest_string(stored_digest))
# 
# Test Cases
# 

def default_toc_checksum_validity(filename):
	with util.archive_created(filename, "/bin") as path:
		_verify_header_checksum(path, "sha1")

def sha1_toc_checksum_validity(filename):
	with util.archive_created(filename, "/bin", "--toc-cksum", "sha1") as path:
		_verify_header_checksum(path, "sha1")

def sha256_toc_checksum_validity(filename):
	with util.archive_created(filename, "/bin", "--toc-cksum", "sha256") as path:
		_verify_header_checksum(path, "sha256")

def sha512_toc_checksum_validity(filename):
	with util.archive_created(filename, "/bin", "--toc-cksum", "sha512") as path:
		_verify_header_checksum(path, "sha512")

def broken_toc_default_checksum(filename):
	with util.archive_created(filename, "/bin") as path:
		# Mess up the archive
		toc_start = _get_header_size(path)
		_clobber_bytes_at(range(toc_start + 4, toc_start + 4 + 100), path) # Why did the original test specify 4? No idea.
		
		# Try to extract it
		_verify_extraction_failed(filename)
		
def broken_toc_sha1_checksum(filename):
	with util.archive_created(filename, "/bin", "--toc-cksum", "sha1") as path:
		# Mess up the archive
		toc_start = _get_header_size(path)
		_clobber_bytes_at(range(toc_start + 4, toc_start + 4 + 100), path) # Why did the original test specify 4? No idea.
		
		# Try to extract it
		_verify_extraction_failed(filename)

def broken_toc_sha256_checksum(filename):
	with util.archive_created(filename, "/bin", "--toc-cksum", "sha256") as path:
		# Mess up the archive
		toc_start = _get_header_size(path)
		_clobber_bytes_at(range(toc_start + 4, toc_start + 4 + 100), path) # Why did the original test specify 4? No idea.

		# Try to extract it
		_verify_extraction_failed(filename)

def broken_toc_sha512_checksum(filename):
	with util.archive_created(filename, "/bin", "--toc-cksum", "sha512") as path:
		# Mess up the archive
		toc_start = _get_header_size(path)
		_clobber_bytes_at(range(toc_start + 4, toc_start + 4 + 100), path) # Why did the original test specify 4? No idea.

		# Try to extract it
		_verify_extraction_failed(filename)

def broken_heap_default_checksum(filename):
	with util.archive_created(filename, "/bin") as path:
		# Mess up the archive
		toc_start = _get_header_size(path)
		toc_size = _get_toc_size(path)
		# Why 32? That's the size of the default sha256 checksum, which is stored before the heap.
		_clobber_bytes_at(range(toc_start + toc_size + 32, toc_start + toc_size + 32 + 100), path)
		
		# Try to extract it
		_verify_extraction_failed(filename)

def default_checksum_algorithm(filename):
	with util.archive_created(filename, "/bin") as path:
		header = subprocess.check_output(["xar", "--dump-header", "-f", path])
		found = False
		for line in header.splitlines():
			matchdata = re.match("^Checksum algorithm:\s+(\d+)\s+\\((\w+)\\)$", line)
			if not matchdata:
				continue
			found = True
			algorithm = matchdata.groups()[1]
			assert algorithm == "SHA1", "unexpected checksum algorithm default: received {a}, expected SHA1".format(a=algorithm)
		assert found, "unexpected output from `xar --dump-header`:\n{h}".format(h=header)


# Apparently, xar doesn't currently fail when given an invalid checksum algorithm. Something to fix later.
# 
# def invalid_checksum_algorithm(filename):
# 	try:
# 		with util.archive_created(filename, "/bin", "--toc-cksum", "invalid-algorithm") as path:
# 			raise AssertionError("xar succeeded when it should have failed")
# 	except subprocess.CalledProcessError:
# 		pass

# It does fail for md5 explicitly, however
def md5_toc_checksum_failure(filename):
	try:
		with open("/dev/null", "a") as devnull:
			with util.archive_created(filename, "/bin", "--toc-cksum", "md5", stderr=devnull) as path:
				raise AssertionError("xar succeeded when it should have failed")
	except subprocess.CalledProcessError:
		pass

def md5_file_checksum_failure(filename):
	try:
		with open("/dev/null", "a") as devnull:
			with util.archive_created(filename, "/bin", "--file-cksum", "md5", stderr=devnull) as path:
				raise AssertionError("xar succeeded when it should have failed")
	except subprocess.CalledProcessError:
		pass

def _verify_checksum_algorithm(filename, algorithm):
	additional_args = []
	if algorithm:
		additional_args = ["--file-cksum", algorithm]
	else:
		algorithm = "sha1"
	
	with util.archive_created(filename, "/bin", *additional_args) as path:
		toc = subprocess.check_output(["xar", "--dump-toc=-", "-f", path])
		found = False
		for line in toc.splitlines():
			if '<unarchived-checksum style="{a}">'.format(a=algorithm) in line or '<archived-checksum style="{a}">'.format(a=algorithm) in line:
				break
		else:
			raise AssertionError("unexpected output from `xar --dump-toc=-`:\n{t}".format(t=toc))

def default_file_checksum_algorithm(filename):
	_verify_checksum_algorithm(filename, None)

def sha1_file_checksum_algorithm(filename):
	_verify_checksum_algorithm(filename, "sha1")

def sha256_file_checksum_algorithm(filename):
	_verify_checksum_algorithm(filename, "sha256")

def sha512_file_checksum_algorithm(filename):
	_verify_checksum_algorithm(filename, "sha512")


TEST_CASES = (default_toc_checksum_validity, sha1_toc_checksum_validity, sha256_toc_checksum_validity, sha512_toc_checksum_validity,
              broken_toc_default_checksum, broken_toc_sha1_checksum, broken_toc_sha256_checksum, broken_toc_sha512_checksum,
              broken_heap_default_checksum, default_checksum_algorithm,
              default_file_checksum_algorithm, sha1_file_checksum_algorithm, sha256_file_checksum_algorithm, sha512_file_checksum_algorithm,
              md5_toc_checksum_failure, md5_file_checksum_failure,)

if __name__ == "__main__":
	for case in TEST_CASES:
		try:
			case("{f}.xar".format(f=case.func_name))
			print("PASSED: {f}".format(f=case.func_name))
		except (AssertionError, IOError, subprocess.CalledProcessError):
			import sys, os
			print("FAILED: {f}".format(f=case.func_name))
			sys.excepthook(*sys.exc_info())
			print("")
